import netgen.gui
import matplotlib.pyplot as plt
from make_1d_mesh import *

import numpy as np
import matplotlib as mpl
from ngsolve import *
# mesh and spaces initialization
def mesh_and_spaces(nel,length,ord):
    m2 , n = SegMesh_refined(nel,0.0,length,1200)     
    m2.dim = 1
    m1=SegMesh(nel,0,length,0)
    m1.dim=1
    mesh = Mesh(m2)
    Y = H1(mesh, order=ord, dirichlet=[1,2])  # space can be fixed for the whole routine
    X = FESpace([Y,Y,Y],flags = { "dgjumps" : True })
    xpoints = np.linspace(0,length,nel+1)
    return X,Y, mesh, xpoints
# scaling parameters
def scaling():
    lam= 0.0433
    beta= 3.86
    z1= 1
    z2= -1
    kappa = 6
    D1=1.33*pow(10,-9)
    D2= 2.07*pow(10,-9)
    return lam, beta, kappa, z1,z2,D1,D2


def PNP_solver(Vapl,bc1,mesh,length,X,Y,Sig,radius,xpoints,nel):
    inform  = 0
    tau = 1
    # shape function prep
    xpoints = GridFunction(Y)
    xpoints.Set(x)
    R = GridFunction(Y)
    R.Set(radius)
    #  surface charge function prep
    Sigma= GridFunction(Y)
    Sigma.Set(Sig)
    lamb, beta,kappa,z1,z2,D1,D2 = scaling()
    A= GridFunction(Y)
    A = R*R*22/7
    # trial and test functions
    V,u1,u2 = X.TrialFunction()
    v1,v2,v3 = X.TestFunction()

    # DG jumps for trial functions
    V_jump = V-V.Other()
    u1_jump = u1-u1.Other()
    u2_jump = u2-u2.Other()
    # DG jumps for test functions
    v1_jump = v1-v1.Other()
    v2_jump = v2-v2.Other()
    v3_jump = v3-v3.Other()
    # normal
    n = specialcf.normal(1)
    #mean gradients
    mean_dVdn = 0.5*n * (grad(V)+grad(V.Other()))
    mean_dc1dn = 0.5*n * (grad(u1)+grad(u1.Other()))
    mean_dc2dn = 0.5*n * (grad(u2)+grad(u2.Other()))

    #mean gradients
    mean_dv1dn = 0.5*n * (grad(v1)+grad(v1.Other()))
    mean_dv2dn = 0.5*n * (grad(v2)+grad(v2.Other()))
    mean_dv3dn = 0.5*n * (grad(v3)+grad(v3.Other()))


    alpha = 4
    order =1
    h = specialcf.mesh_size

    # bilinear form
    b = CoefficientFunction( (x) )
    bn = b*specialcf.normal(1)
    a = BilinearForm(X)
    a += SymbolicBFI( lamb*A*grad(V) * grad(v1) - A*kappa*z1*exp(u1 - z1*beta*V)* v1 \
                     - A*kappa*z2* exp(u2 - z2*beta*V)*v1  -2*22/7*R*Sigma*v1) # poisson Boltzamann equation
    a += SymbolicBFI(-mean_dVdn*v1_jump -mean_dv1dn*V_jump, element_boundary=True)  #  This is the first place I wanted to implement jumps
    a += SymbolicBFI(- A*exp(u1 - z1*beta*V)*grad (u1)* grad(v2)) # first cont equation
    a += SymbolicBFI(- A*exp(u2 - z2*beta*V)*grad (u2)* grad(v3))# second cont equation
    
    
    
    c = Preconditioner(a, "multigrid")
    # boundary conditions  and initial guesses 
    u = GridFunction(X)
    u.components[0].Set(x/length*Vapl)
    u.components[1].Set(z1*beta*u.components[0]  + log(bc1))
    u.components[2].Set(z2*beta*u.components[0]  + log(bc1))
    res = u.vec.CreateVector()
    du = u.vec.CreateVector()
    ################################# newton loop
    for it in range(250):      
        a.Apply(u.vec, res)
        a.AssembleLinearization(u.vec)        
        du.data = a.mat.Inverse(X.FreeDofs()) * res
        u.vec.data -= du
        #stopping criteria
        stopcritval = sqrt(abs(InnerProduct(du,res)))
        if inform ==1:
            print ("Iteration",it)
            print ("<A u",it,", A u",it,">_{-1}^0.5 = ", stopcritval)
         
        if stopcritval < 1e-10:
            break  
    ################################## current calculations  and output functions    
    c1 = GridFunction(Y)
    c2 = GridFunction(Y)
    Vout = GridFunction(Y)
    Vout.Set(u.components[0])
    u1= GridFunction(Y)
    u1.Set(u.components[1])
    u2= GridFunction(Y)
    u2.Set(u.components[2])
    c1.Set(exp(u.components[1] - z1*beta*u.components[0]))
    c2.Set(exp(u.components[2] - z2*beta*u.components[0]))
    gradu1= grad(u1)
    gradu2= grad(u2)
    J1 = GridFunction(Y)
    J2 = GridFunction(Y)
    J1.Set(D1*A*c1*gradu1)
    J2.Set(D2*A*c2*gradu2)
    e = 1.602*pow(10,-9)
    I = pow(10,2)*(-z1*J1.vec[int(nel/2)] - z2*J2.vec[int(nel/2)])
    #plt.plot(xpoints.vec,Vout.vec)
    #plt.show()
    return Vout, c1, c2, J1.vec[int(nel/2)], J2.vec[int(nel/2)]


# Make 1D mesh
nel = 500
length = 3600
Vap = 2
bc = 1
tau = 20 #  for  steepest fradient 
max_back = 10 
ngsglobals.msg_level=1
X,Y,mesh,xpoints = mesh_and_spaces(nel,length,1)
# domain parameters
xpoints = GridFunction(Y)
xpoints.Set(x)
Sigma= GridFunction(Y)
R = GridFunction(Y)
R.Set(x) # usually a discontinous function  -> here  I made it simpler 
Sigma.Set(x) # usually a discontinous function -> here  I made it simpler 


Vp,c1p,c2p,J1p,J2p = PNP_solver(Vap,bc,mesh,length,X,Y,Sigma,R,xpoints,nel)
Vm,c1m,c2m,J1m,J2m = PNP_solver(-Vap,bc,mesh,length,X,Y,Sigma,R,xpoints,nel)
